// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __grammar_h
#define __grammar_h

#include "c_all_internal.h"
#include "hashmap.h"
#include "sortedset.h"
#include "intset.h"

#define PROP_FLAGS_REJECT               (1)
#define PROP_FLAGS_RANK                 (2)
#define PROP_FLAGS_ASSOC                (4)
#define PROP_FLAGS_LABEL                (8)
#define PROP_FLAGS_ACTION               (16)
#define PROP_FLAGS_RUNTIME_RESOLVABLE   (32)

class Symbol;
class SymbolSet;
class Production;
class Grammar;
class TokenSetMergeAlgorithm;
class LRItem;
class LRStateSet;
class ProductionLabel;
class OrderedTokenSet;
class Literal;
class Regex;
class CharSet;
class Nonterminal;
class Terminal;
class Rx;
class OptionalRange;
class AllProperties;
class Nfa;
class FrozenDfa;
class TokenSetActions;

enum AType {ATypeNone=0,ATypeShift=1,ATypeReduce=2,ATypeAccept=4,ATypeReject=8};
enum SType { LITERAL_SYMBOL, REGEX_SYMBOL, CHARSET_SYMBOL, NONTERMINAL_SYMBOL, TERMINAL_SYMBOL };
enum sl_decl { sl_neutral, sl_short, sl_long };
enum GType {GTypeNone, GTypeSymbol, GTypeProduction};
enum GrammarSolution
{
    DeterministicParser = 0,
    NondeterministicParser = 1,
    NondeterministicParser_ScannerLess = 2,
    UseBestAchievableSolution = 3,
};

class LIBGLR_EXPORT CodeSnip
{
public:
    CodeSnip();
    CodeSnip(String file, int line, String code);
    ~CodeSnip();
    String m_file;
    int m_line;
    String m_code;
    int count();
    void clear();
};

class LIBGLR_EXPORT LRItem_eq_Exact
{
public:
    bool operator()(LRItem* const &lhs, LRItem* const &rhs) const;
};

class LIBGLR_EXPORT LRItem_hash_Exact
{
public:
    uint32_t operator()(LRItem* const &p) const;
};

typedef hashset<LRItem*,16,LRItem_eq_Exact,LRItem_hash_Exact> lritem_ptrset;

class LIBGLR_EXPORT SymbolSet
{
public:
    UIntSet m_chars;
	sortedset<Symbol*> m_symbols;
public:
    SymbolSet();
    SymbolSet(const SymbolSet &rhs);
    SymbolSet &operator=(const SymbolSet &rhs);
    void operator+=(const SymbolSet &rhs);
    void operator+=(const UIntSet &rhs);
    void operator+=(Symbol *rhs);
    void operator-=(const SymbolSet &rhs);
    void operator-=(const UIntSet &rhs);
    void operator-=(Symbol *rhs);
    bool operator&(unsigned int rhs) const;
    bool operator&(const Symbol *rhs) const;
    void operator&=(const SymbolSet &rhs);
    bool operator==(const SymbolSet &rhs) const;
    bool operator!=(const SymbolSet &rhs) const;
	SymbolSet operator-(const SymbolSet &rhs) const;
	SymbolSet operator&(const SymbolSet &rhs) const;
    void clear();
    int count();
    void Print(outbuf *p) const;
    void FlattenCharsets();
};

class LIBGLR_EXPORT LRItem
{
public:
    LRItem() : m_ss(0), m_p(0), m_pos(-1) {}
    LRItem(LRStateSet *ss, Production *p, int pos) : m_ss(ss), m_p(p), m_pos(pos) {}
    LRStateSet *m_ss;
    Production *m_p;
    int m_pos;
    SymbolSet m_follows;
    Symbol *GetSymbolAtPosition() const;
    SymbolSet GetFollowsAtPosition();
    bool ChildAllowedAtPosition(Production *childproduction) const;
    void Print(outbuf *pio) const;
    bool operator==(const LRItem &rhs) const;
    bool operator!=(const LRItem &rhs) const { return ! (*this == rhs); }
    bool operator<(const LRItem &rhs) const;
    static int cmp_lritem(const LRItem *lhs, const LRItem *rhs);
    lritem_ptrset GetPostReduceItems() const;
    lritem_ptrset GetPostReduceItems_Verbose(outbuf *verbose) const;
    lritem_ptrset GetReducibleItems() const;
protected:
    lritem_ptrset GetPrevItems();
    lritem_ptrset GetNextItems(GType gtype, Symbol *s, Production *p);
};

class LIBGLR_EXPORT GotoItem
{
public:
    GotoItem();
    GotoItem(GType gtype, Symbol *s, Production *p, LRStateSet *to);
    GType m_gtype;
    Symbol *m_s;
    Production *m_p;
    LRStateSet *m_to;
};

class LIBGLR_EXPORT LRAction
{
public:
    LRStateSet *m_state;
    GType m_gtype;
    Symbol *m_s;
    Production *m_p;
    AType m_atype;
    Production *m_reduceby;
    int m_itemidx;
    LRStateSet *m_shiftto;
    bool operator==(const LRAction &rhs) const;
    bool operator<(const LRAction &rhs) const;
    void Print(outbuf *pio) const;
    const char *ActionType() const;
    bool IsRuntimeResolvable() const;
    int symbolidx() const;
    int actionidx() const;
    static const char *static_ActionType(AType atype);
    static int cmp_lraction(const LRAction *lhs, const LRAction *rhs);
    static int cmp_actionsymbol(const LRAction *lhs, const LRAction *rhs);
};

class LIBGLR_EXPORT assocmapkey
{
public:
    const Production *m_parent;
    const Production *m_child;

    assocmapkey(const assocmapkey &rhs);
    assocmapkey(const Production *parent, const Production *child);
    bool operator==(const assocmapkey &rhs) const;
    bool operator!=(const assocmapkey &rhs) const;
};

template<>
class simpleeq <assocmapkey>
{
public:
    bool operator()(const assocmapkey &lhs, const assocmapkey &rhs) const { return lhs == rhs; }
};

template<>
class simplehash <assocmapkey>
{
public:
    uint32_t operator()(const assocmapkey &v) const
    {
        uint32_t u2[2] = {(uint32_t)(intptr_t)v.m_parent, (uint32_t)(intptr_t)v.m_child};
        return hash_uint32s(u2, 2, 101);
    }
};

// Assoc describes an association, that is, where a child may or may not nest in a parent production.
class LIBGLR_EXPORT Assoc
{
public:
    enum Association { None, Left, Right, Non, ByPosition, ByAbsPosition };
    enum AssocStyle  { NoStyle, Forbid, Allow, AllowAny, ForbidAny };
    Assoc();
    AssocStyle m_assocstyle;
    Production *m_parent;
    Production *m_child;
    UIntSet m_assoc_pos;
    vector<LRItem> m_from_series;
    String m_fname;
    int m_lineno;
    bool operator==(const Assoc &rhs) const;
    bool operator<(const Assoc &rhs) const;
    LRItem GetFrom(int pos) const;
};

// AssocItem is a statekeeper within LRStateSet for tracking which association rules are active.
// Not all AssocItems are active all the time.  When an AssocItem's position marker (m_pos) reaches
// the end of the Assoc's from_series, then it becomes active, and the rule that the Assoc class
// describes must be enforced (see Production::ChildAllowedAtPosition).  Until then, it is just part
// of the state that helps to distinguish it from any other state.
class LIBGLR_EXPORT AssocItem
{
public:
    AssocItem();
    AssocItem(const Assoc &assoc, size_t pos);
    Assoc m_assoc;
    size_t m_pos;
    bool IsActive() const;
    LRItem NextFrom() const;
    bool operator==(const AssocItem &rhs) const;
    bool operator<(const AssocItem &rhs) const;

    void Print(outbuf *pio) const;
};

class LIBGLR_EXPORT AssocItems
{
public:
    sortedset<AssocItem> m_active_allow;
    sortedset<AssocItem> m_inactive_allow;
    sortedset<AssocItem> m_active_forbid;
    sortedset<AssocItem> m_inactive_forbid;

    bool HasActiveAssocItem(Assoc::AssocStyle assocstyle, int position) const;
    void AddAssocItem(const AssocItem &associtem);

    bool operator==(const AssocItems &rhs) const;
    bool operator!=(const AssocItems &rhs) const;
};

class LIBGLR_EXPORT Assocs
{
public:
    vector<Assoc> m_1st_order_allow;        // always on
    vector<Assoc> m_1st_order_forbid;       // always on
    vector<Assoc> m_gt_1st_order_allow;
    vector<Assoc> m_gt_1st_order_forbid;

    vector<Assoc>::iterator FindOrCreateAssociation(Production *parent, Production *child, Assoc::AssocStyle assocstyle, const vector<LRItem> &from_series, String fname, int lineno);
    bool Has1stOrderAssoc(Assoc::AssocStyle assocstyle, int position) const;
};

typedef hashmap<assocmapkey, AssocItems> associtem_map;
typedef hashmap<assocmapkey, Assocs > assoc_map;
typedef hashmap<UIntSet, int> rangemap;

struct LRActionRange
{
    LRStateSet *m_state;
    AType m_atype;
    int m_firstactionidx;
    int m_lastactionidx;
    int m_rangeidx;
};

class LIBGLR_EXPORT LRStateSet
{
public:
    Grammar *m_g;
    LRStateSet(Grammar *g);
    int m_number; // The state number
    vector<LRItem> m_items;
    vector<GotoItem> m_gotos;
    vector<GotoItem> m_backgotos;
    vector<LRAction> m_actions;
    vector<LRActionRange> m_actionranges;
    associtem_map m_associtems;

    bool operator==(const LRStateSet &rhs) const;
    void Print(outbuf *pio) const;
    
    // LRItem functions
    void AddLRItem(Production *p, int position);
    void AddGotoItem(GType gtype, Symbol *s, Production *p, LRStateSet *to);

    // LRAction functions
    void AddReduceAction(Production *reduceby, int itemidx, Symbol *s, AType atype);
    void AddShiftAction(GType gtype, Symbol *symbol, Production *production, LRStateSet *shiftto);
    void SLRClosure(hashset<Production*> &addedproductions);
    void SLRClosure_Verbose(hashset<Production*> &addedproductions, outbuf *verbose);
    void GenerateActions(rangemap &ranges, vector<UIntSet> &rangesbyidx);

    // AssocItem functions
    const AssocItems *GetAssocItems(const Production *parent, const Production *child) const;
};

class LIBGLR_EXPORT PriorityGroup
{
public:
    Assoc::Association m_assoc;
    vector<IntSetRange> m_assoc_pos;
    vector<ProductionLabel*> m_labels;
    vector<LRItem> m_from_series;
    PriorityGroup();
    PriorityGroup(Assoc::Association Assoc);
    PriorityGroup(Assoc::Association Assoc, ProductionLabel *first, ...); // NULL Terminated
};

class LIBGLR_EXPORT PrioritySet
{
public:
    vector<PriorityGroup> m_groups;
    hashmap<ProductionLabel*,int,64> m_grouppriorities;
    String m_fname;
    int m_lineno;
    PrioritySet();
    PrioritySet &operator=(const PrioritySet &rhs);
    PrioritySet(String fname, int lineno, vector<PriorityGroup> &groups);
    int GetPriorityIndex(ProductionLabel *label);
    void FixatePriorityIndices();
};

class LIBGLR_EXPORT ProductionLabel
{
public:
    Grammar *m_g;
    String m_Name;
    ProductionLabel(Grammar *g, String Name);
    void Print(outbuf *pio) const;
    int GetIndex();
};

class LIBGLR_EXPORT ActionConflict
{
public:
    ActionConflict()
        : m_state(0), m_action(ATypeNone), m_otheraction(ATypeNone), m_gtype(GTypeNone), m_p(0), m_reducebyidx(-1), m_otherreducebyidx(-1)
    {
    }
    ActionConflict(LRStateSet *state, AType action, AType otheraction, GType gtype, SymbolSet set, Production *p, int reducebyidx, int otherreducebyidx)
        : m_state(state), m_action(action), m_otheraction(otheraction), m_gtype(gtype), m_set(set), m_p(p), m_reducebyidx(reducebyidx), m_otherreducebyidx(otherreducebyidx)
    {
    }

    LRStateSet *m_state;
    AType m_action;
    AType m_otheraction;
    GType m_gtype;
    SymbolSet m_set;
    Production *m_p;
    int m_reducebyidx;
    int m_otherreducebyidx;
    const char *ActionType() const { return LRAction::static_ActionType(m_action); }
    const char *OtherActionType() const { return LRAction::static_ActionType(m_otheraction); }
};

class LIBGLR_EXPORT NamedCharSet
{
public:
    String m_name;

private:
    UIntSet m_chars;

public:
    NamedCharSet();
    NamedCharSet(String name, const UIntSet &charset);
    UIntSet chars();
};

class LIBGLR_EXPORT ProductionOrLabel
{
public:
    ProductionOrLabel();
    ProductionOrLabel(Production *p);
    ProductionOrLabel(ProductionLabel *l);
    enum ProductionGroupStyle { IsNone, IsProduction, IsLabel, IsAny };
    ProductionGroupStyle m_ProductionGroupStyle;
    ProductionLabel *m_l;
    Production *m_p;

    vector<Production*> GetProductions(Grammar *g);
};

class LIBGLR_EXPORT AssocByLabel
{
public:
	AssocByLabel();
    AssocByLabel(ProductionOrLabel parentlabel, ProductionOrLabel childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &assoc_pos, const vector<LRItem> &from_series, const String fname, int lineno);
    Assoc::AssocStyle m_assocstyle;
	Assoc::Association m_assoc;
    ProductionOrLabel m_parentlabel;
    ProductionOrLabel m_childlabel;
    vector<IntSetRange> m_assoc_pos;
    vector<LRItem> m_from_series;
	String m_fname;
	int m_lineno;
};

class LIBGLR_EXPORT SkipRange
{
public:
    SkipRange() : m_start(-1), m_stop(-1) {}
    SkipRange(int start, int stop) : m_start(start), m_stop(stop) {}
    SkipRange(const OptionalRange &rhs);
    int m_start;
    int m_stop;
};

class LIBGLR_EXPORT OptionalRange
{
public:
    OptionalRange();
    OptionalRange(int start, int stop);
    int m_start;
    int m_stop;
    vector<OptionalRange> m_optionalranges;
    static vector< vector<SkipRange> > GetListPermutations(const vector<OptionalRange> &list);
    vector< vector<SkipRange> > GetPermutations() const;
    bool operator==(const OptionalRange &rhs);
    bool operator!=(const OptionalRange &rhs);
};

class rule
{
public:
    int ruleno;
    String displaystring;
};

class conflict
{
public:
    int state;
    int action;
    int otheraction;
    int reduceidx;
    int otherreduceidx;
    String symbol;
    int resolved;
};

class stateitem
{
public:
    int rowid;
    int state;
    int item;
    String displaystring;
    String allfollows;
};

class dep
{
public:
    int dstid;
    int srcid;
};

class follow
{
public:
    int dstid;
    int srcid;
    int isdep;
    String symbol;
};

class LIBGLR_EXPORT verboseinfo
{
public:
    vector<rule> m_rules;
    vector<conflict> m_conflicts;
    vector<stateitem> m_stateitems;
    vector<dep> m_deps;
    vector<follow> m_follows;

    void clear();
    int AddItem(int stateno, int itemidx, const char *displaystring, const char *allfollows);
    void UpdateItem(int rowid, const char *allfollows);
    void AddDep(int dstid, int srcid);
    void AddFollow(int dstid, int srcid, int isdep, const char *symbol);
};

class LIBGLR_EXPORT defines
{
private:
    hashmap<String,int> m_defcnt;
    vector< hashmap<String,int> > m_scopes;

public:
    void define(const char *s);
    void undefine(const char *s);
    int is_defined(String &s);
    void push();
    void pop();
};

class LIBGLR_EXPORT emitoptions
{
public:
    String m_skeletonfile;
    defines defined_symbols;
};

class LIBGLR_EXPORT Grammar
{
public:
    // For access only!  Don't modify the vectors themselves.
    vector<Production*> m_productions;
    typedef hashmap< Production*, int > indexmap;
    typedef hashmap< Symbol*, int > symbolidmap;
    indexmap m_indexmap;
    symbolidmap m_symbolidmap;
    vector<Symbol*> m_symbols;
    vector<NamedCharSet> m_namedcharsets;
    LRStateSet *m_initialstate;
    vector<LRStateSet*> m_lrstates;
    vector<ProductionLabel*> m_labels;
    vector<OrderedTokenSet*> m_TokenSets;
    // Filled in after solving.
    Symbol *m_endmarker;
    Symbol *m_startsymbol;
    Symbol *m_errorsymbol;
    bool m_bUsePoundLine;
    vector<ActionConflict> m_ActionConflicts;
    vector<ActionConflict> m_ActionResolvedConflicts;
    String m_name;
    int m_errcount;
    int m_wrncount;

public:
    bool m_bLogParse;
    bool m_bLogFirsts;
    bool m_bLogFollows;
    bool m_bLogClosure;
    // borrowed, not owned {
    outbuf *m_err;
    outbuf *m_verbose;
    verboseinfo *m_verbosedb;
    // }
    vector<UIntSet> m_rangesbyidx;

private:
    // main copy in m_symbols {
    vector<Literal*> m_literals;
    vector<Regex*> m_regexs;
    vector<CharSet*> m_charsets;
    vector<Nonterminal*> m_nonterminals;
    vector<Terminal*> m_terminals;
    // m_symbols }
    vector<PrioritySet> m_priorities;
	vector<AssocByLabel> m_assocbylabels;
    int m_ntcounter;
    TokenSetMergeAlgorithm *m_TokenSetMerge;
    assoc_map m_assocs;
    typedef hashmap< String, Rx* > tokfragmap;
    tokfragmap m_tokfragmap;
    rangemap m_ranges;

protected:
    void ApplyAssociations();
    void ApplyAssociationsByLabel(ProductionOrLabel parentlabel, ProductionOrLabel childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, vector<IntSetRange> &assoc_pos, const vector<LRItem> &from_series, String fname, int lineno);
    void AddStartState();
    void CheckForUnreachableAndProductionlessNonterminals();
    void RemoveUnusedParts();
    void ExpandTokensIntoProductions();
    void GenerateOptionalPartPermutations();
    void BuildStatesAndGotos();
    void ComputeFirsts();
    void PrintFirsts(outbuf *p) const;
    void FixupFollowRestricts();
    void PrintFollowRestricts(outbuf *p) const;
    void ComputeFollows();
    void ComputeFollows_Verbose();
    void PrintFollows(outbuf *p) const;
    void GenerateActions();
    void CollectConflicts();
    void RemoveProduction(int index);
    void TokenStateMissReport(outbuf *pio) const;
    Production *AddProductionCore(Symbol *nt, vector<Symbol*> &symbols, vector<sl_decl> &sldecls, vector<OptionalRange> &optionalranges, AllProperties *props);
    void GetSymbolsAndRanges(vector<Symbol*> &callsymbols, vector<sl_decl> &callsldecls, vector<Symbol*> &symbols, vector<sl_decl> &sldecls, vector<OptionalRange> &optionalranges);
    void UpdateRuleSet(LRStateSet &tmplrstateset, LRStateSet *lrss);

public:
    GrammarSolution m_RequestedGrammarSolution;
    void SetOption(const char *option, const char *op, const char *value);
    bool SolvingForScannerless() { return m_RequestedGrammarSolution == NondeterministicParser_ScannerLess; }
    Symbol *m_epsilon;
    CodeSnip m_StartBlock;
    CodeSnip m_EndBlock;
    String m_globalctype;
	String m_fname_root;
    String m_logfile;
    String m_outfile;
    OrderedTokenSet *m_InitialTokenState;
    String m_EmitLanguage;
	bool m_bLogDfa;

    // Constructor
    Grammar();
    virtual ~Grammar();

    // Symbol Add/Find/Remove functions
    Nonterminal *AddAnonymousNonterminal(OrderedTokenSet *tokstate = 0);
    Nonterminal *FindOrAddNonterminal(String name, OrderedTokenSet *tokstate = 0, bool bCreateIfNotFound = true);
    Literal *FindOrAddLiteral(String name, bool case_insensitive, OrderedTokenSet *tokstate = 0, bool bCreateIfNotFound = true);
    Regex *FindOrAddRx(String name, Rx *rx, bool case_insensitive, OrderedTokenSet *tokstate = 0, bool bCreateIfNotFound = true);
    Terminal *FindOrAddTerminal(String name);
    CharSet *FindOrAddCharset(UIntSet chars);
    void RemoveSymbol(Symbol *s);
    UIntSet LookupNamedCharSet(String name, bool *pbFound = 0);
    bool DeclareTokFrag(String name, Rx *rx);
    Rx *LookupTokFrag(String name);

    // Production functions
    Production *AddProduction(Symbol *nt, vector<Symbol*> &symbols, vector<sl_decl> &sldecls, AllProperties *props);
    Production *FindOrAddProduction(Symbol *nt, vector<Symbol*> &symbols, vector<sl_decl> &sldecls, bool &bNew);
    Production *AddSimpleProduction(Symbol *nt, ...); // last parameter must be 0
	Production *vAddSimpleProduction(Symbol *nt, va_list args);
    int GetProductionIndex(Production *p);
    int GetSymbolIndex(Symbol *s);
    int GetProductionLabelIndex(ProductionLabel *pl);
    int GetOrderedTokenSetIndex(OrderedTokenSet *ots);
    void MoveProduction(Production *p, int newindex);

    // LRStateSet functions
    LRStateSet *AddLRStateSet();
    LRStateSet *AddLRStateSet(const LRStateSet &copy);
    LRStateSet *FindLRStateSet(const LRStateSet &match);

    // Label functions
    ProductionLabel *FindOrAddLabel(String label);
    
    // Priority (precedence) functions
    void AddPriorities(PrioritySet &priorities);
    void AddAssociationsByLabel(ProductionOrLabel parentlabel, ProductionOrLabel childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &forbid_pos, const vector<LRItem> &from_series, const String fname, int lineno);
    bool TestForPriorityConflict(ProductionLabel *parentlabel, ProductionLabel *childlabel);
    
    // assoc functions
    vector<Assoc>::iterator FindOrCreateAssociation(Production *parent, Production *child, Assoc::AssocStyle assocstyle, const vector<LRItem> &from_series, String fname, int lineno);
    const Assocs *GetAssocs(const Production *parent, const Production *child) const;

    // Follow restrict functions
    bool AddFollowRestriction(Symbol *nt, Symbol *adding);

    // TokenSet functions
    OrderedTokenSet *FindTokenSet(const String name);
    OrderedTokenSet *FindOrAddTokenSet(const String name);
    void AddDfa(Nfa &dfa, vector<Symbol*> &accepttokens, OrderedTokenSet *tokenset);

    int Parse(const char *fname);
    // Grammar solving functions
    bool Solve(outbuf *log);
    bool Emit(emitoptions &options);
    
    // Printing functions
    void PrintProductions(outbuf *pio) const;
private:
    void PrintLog(outbuf *pio) const;
};

class LIBGLR_EXPORT TokenSetAction
{
public:
    enum TokenSetActionType {NOACTION, GOTO, PUSH, POP, IGNOREWS};
    TokenSetAction();
    TokenSetAction(String file, int lineno, Symbol *sym, TokenSetActionType action, OrderedTokenSet *actionset);
	String m_file;
    int m_lineno;
    Symbol *m_sym;
    TokenSetActionType m_action;
    OrderedTokenSet *m_actionset;
    
    bool operator==(const TokenSetAction &rhs) const;
    bool operator!=(const TokenSetAction &rhs) const { return ! operator==(rhs); }
    bool ApplyTo(Grammar *g, vector<OrderedTokenSet*> &statestack, outbuf *err); // throws MergeTokenSetStatesException
    void MergeTo(TokenSetActions &tsa); // throws MergeTokenSetStatesException
    void Print(outbuf *pio) const;
};

class LIBGLR_EXPORT TokenSetActions
{
public:
    vector<TokenSetAction> m_actions;
    bool operator==(const TokenSetActions &rhs) const { return m_actions == rhs.m_actions; }
    bool operator!=(const TokenSetActions &rhs) const { return m_actions != rhs.m_actions; }
    TokenSetAction *FindTokenSetAction(Symbol *forsymbol);
    void ApplyTo(Grammar *g, vector<OrderedTokenSet*> &statestack, outbuf *err); // throws MergeTokenSetStatesException
    void MergeTo(TokenSetActions &tsa); // throws MergeTokenSetStatesException
    void Print(outbuf *pio) const;
    vector<TokenSetAction>::iterator begin() { return m_actions.begin(); }
    vector<TokenSetAction>::iterator end()   { return m_actions.end(); }
};

struct compressed_lxitem
{
    int pos;
    int off;
    int count;
    int lead;
    int next;
};

class LIBGLR_EXPORT OrderedTokenSet
{
public:
    Grammar *m_g;
    String m_Name;
	bool m_bHideState;
    vector<Symbol*> m_symbols;
    TokenSetActions m_actions;

    // Outputs of GenerateTokens.
    Nfa *m_ubigdfa;
    FrozenDfa *m_bigdfa;
    vector<int> m_compressed_dfa_toksetbreaks;
    vector<compressed_lxitem> m_compressed_dfa_data;
    vector<Symbol*> m_accepttokens;

public:
	OrderedTokenSet(Grammar *g);
    vector<TokenSetAction>::iterator beginaction() { return m_actions.begin(); }
    vector<TokenSetAction>::iterator endaction()   { return m_actions.end(); }
    TokenSetAction *FindTokenSetAction(Symbol *forsymbol) { return m_actions.FindTokenSetAction(forsymbol); }
    void Print(outbuf *pio) const;
    void vAddTokensAndRules(Symbol *s, TokenSetAction::TokenSetActionType action, OrderedTokenSet *to, const char *file, int lineno, ...);
    void AddTokensAndRules(vector<TokenSetAction> &actions);
    void GenerateTokens();
    void CollectUsedItems(hashset<Production*> &usedproductions, hashset<Symbol*> &usedsymbols, hashset<OrderedTokenSet*> &usedtokensets);
    void AddBigDfaAsCfg();
    int GetIndex();
};

class LIBGLR_EXPORT Symbol
{
public:
	static Symbol *LB();
	static Symbol *RB();
	static Symbol *DOT();
    Grammar *m_g;
    Symbol *m_cloneof;
    vector<Symbol*> m_clones;

public:
    String m_Name;
    String m_CType; // Type within C/C++ Code Snippets
    OrderedTokenSet *m_TokenState;
    vector<Production*> m_OwningProductions; // Productions we appear as nt in.
    SymbolSet m_ntfirsts;
    SymbolSet m_follow_restrict;
    SymbolSet m_firsts;
    bool m_only_child;

    Symbol(Grammar *g, String name, OrderedTokenSet *tokstate);
    virtual void CollectUsedItems(hashset<Production*> &usedproductions, hashset<Symbol*> &usedsymbols, hashset<OrderedTokenSet*> &usedtokensets);
    virtual bool IsTerminal() const { return false; }
    virtual bool IsDeterministicTerminal() const { return false; }
    virtual bool IsCharSet() const { return false; }
    virtual bool IsStartSymbol() const { return false; }
    Symbol *GetCloneParent() { return m_cloneof; }
    virtual void Print(outbuf *pio) const;
    virtual void GenerateToken() {}
    virtual bool AddToNfa(Nfa &nfa) { return false; }
    void AddFollowRestriction(Symbol *s);
	virtual void MarkReachableNonterminalsAndCheckForProductionlessNonterminals(hashset<Nonterminal*> &nonterminals, outbuf *err);
    int GetIndex();
    // These are needed for the current grammar dumping function.
    virtual SType GetSType() const = 0;
    virtual bool IsCaseInsensitive() const { return false; }
	virtual void ResetFirsts();
    virtual bool SetFriendlyName(const char *friendlyname) { return false; };
    virtual Rx *RxFromSymbol();
};

class LIBGLR_EXPORT Literal : public Symbol
{
public:
    bool m_bCaseInsensitive;
    Literal(Grammar *g, String name, OrderedTokenSet *tokstate, bool bCaseInsensitive);
    ~Literal();
    virtual void Print(outbuf *pio) const;
    virtual void GenerateToken();
    virtual bool AddToNfa(Nfa &nfa);
    virtual SType GetSType() const { return LITERAL_SYMBOL; }
    virtual bool IsCaseInsensitive() const { return m_bCaseInsensitive; }
    virtual bool IsDeterministicTerminal() const { return true; }
	virtual void ResetFirsts();
    virtual bool IsTerminal() const ;
    virtual Rx *RxFromSymbol();
};

class LIBGLR_EXPORT Regex : public Symbol
{
public:
    Rx *m_rx;
    bool m_bCaseInsensitive;
    bool m_bAliased;
    Regex(Grammar *g, String name, OrderedTokenSet *tokstate, Rx *rx, bool bCaseInsensitive);
    ~Regex();
    virtual void Print(outbuf *pio) const;
    virtual void GenerateToken();
    virtual bool AddToNfa(Nfa &nfa);
    virtual SType GetSType() const { return REGEX_SYMBOL; }
    virtual bool IsCaseInsensitive() const { return m_bCaseInsensitive; }
    virtual bool IsDeterministicTerminal() const { return true; }
	virtual void ResetFirsts();
    virtual bool IsTerminal() const;
    virtual bool SetFriendlyName(const char *friendlyname);
    virtual Rx *RxFromSymbol();
};

class LIBGLR_EXPORT CharSet : public Symbol
{
public:
    UIntSet m_chars;
    CharSet(Grammar *g, UIntSet chars);
    ~CharSet();
    virtual void Print(outbuf *pio) const;
    virtual bool IsTerminal() const { return true; }
    virtual bool IsDeterministicTerminal() const { return true; }
    virtual bool IsCharSet() const { return true; }
    virtual SType GetSType() const { return CHARSET_SYMBOL; }
	virtual void ResetFirsts();
};

class LIBGLR_EXPORT Nonterminal : public Symbol
{
public:
    Nonterminal(Grammar *g, String name, OrderedTokenSet *tokstate);
    ~Nonterminal();
    virtual void MarkReachableNonterminalsAndCheckForProductionlessNonterminals(hashset<Nonterminal*> &nonterminals, outbuf *err);
    virtual SType GetSType() const { return NONTERMINAL_SYMBOL; }
};

class LIBGLR_EXPORT StartSymbol : public Nonterminal
{
public:
    StartSymbol(Grammar *g);
    ~StartSymbol();
    virtual bool IsStartSymbol() const { return true; }
};

class LIBGLR_EXPORT Terminal : public Symbol
{
public:
    Terminal(Grammar *g, String name);
    ~Terminal();
    virtual bool IsTerminal() const { return true; }
    virtual bool IsDeterministicTerminal() const { return true; }
    virtual SType GetSType() const { return TERMINAL_SYMBOL; }
	virtual void ResetFirsts();
};

void LIBGLR_EXPORT ComputeFirstsFromSymbolString(vector<Symbol*>::const_iterator first, vector<Symbol*>::const_iterator end, SymbolSet &firsts, Symbol *epsilon);

class LIBGLR_EXPORT AllProperties
{
public:
    AllProperties() : m_flags(0), m_rank(0), m_assoc(Assoc::None), m_label(0) {}
    int m_flags;
    int m_rank;
    Assoc::Association m_assoc;
    vector<IntSetRange> m_assoc_pos;
    vector<LRItem> m_from_series;
    ProductionLabel *m_label;
    CodeSnip m_ReduceCode;

    void operator|=(AllProperties &rhs);
};

enum DisambiguationPreferenceType { NoPreference=0,Prefer=1,Avoid=2 };

class LIBGLR_EXPORT Production
{
public:
    Production *m_copyof;
    vector<Production*> m_copycats;
    vector<SkipRange> m_skips;
    bool m_reject;
    vector<OptionalRange> m_optionalranges;

protected:
    friend class Grammar;
    bool m_finalized;

public:
    Grammar *m_g;
    Symbol *m_nt;
    vector<Symbol*> m_items;
    vector<sl_decl> m_sldecls;
    ProductionLabel *m_label;
    CodeSnip m_ReduceCode;
    bool m_runtime_resolvable;
    bool IsRuntimeResolvable();
    // Predeclared items allow you to use productions in %forbid rules before they come up in the grammar.
    // Only once they're actually declared though, do we flip the m_bPredeclareItem flag to true.
    bool m_bPredeclareItem;
    SymbolSet m_follows;
    int m_rank;

    Production(Grammar *g, Symbol *nt, vector<Symbol*> &symbols, vector<sl_decl> &sldecls);
    ~Production();
    void MarkReachableNonterminalsAndCheckForProductionlessNonterminals(hashset<Nonterminal*> &nonterminals, outbuf *err);
    void Print(outbuf *pio, int position = -1) const;
    void ComputeFirsts(SymbolSet &firsts);
    int GrammarIndexToPosition(int index);
    int PositionToAssocPosition(int index);
    bool IsRejectProduction() { return m_reject; }
    int GetIndex();
    vector< vector<SkipRange> > GenerateOptionalPartPermutations();
    void CollectUsedItems(hashset<Production*> &usedproductions, hashset<Symbol*> &usedsymbols, hashset<OrderedTokenSet*> &usedtokensets);
    void SetProps(AllProperties *props);
    void FinalizeDeclaration(int newindex);
    bool AddAssociation(Production *child, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &byposition_positions, const vector<LRItem> &from_series, const String fname, int lineno);
    void AddAssociationByLabel(ProductionLabel *childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &byposition_positions, const vector<LRItem> &from_series, const String fname, int lineno);
    bool ChildAllowedAtPosition(const LRStateSet &lrstate, Production *child, int position) const;
    bool IsErrorProduction();
};

#endif
