// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __hashmap_h
#define __hashmap_h

#include "hashset.h"

template<class T, class eqT>
class keyeq
{
    eqT m_eqT;
public:
    bool operator()(const T &lhs, const T &rhs) const
    {
        return m_eqT(lhs.first, rhs.first);
    }
};

template<class T, class hashT>
class keyhash
{
    hashT m_hashT;
public:
    uint32_t operator()(const T &t) const
    {
        return m_hashT(t.first);
    }
};

template<class Key, class Value, int bucketsize = 512, class eqT=simpleeq<Key>, class hashT=simplehash<Key> >
class hashmap
{
public:
    typedef Key KeyType;
    typedef pair<const Key, Value> kvpair;
    typedef hashcore< kvpair, bucketsize, keyeq<kvpair,eqT>, keyhash<kvpair,hashT> > hcore;
    typedef typename hcore::nodebucket nodebucket;
    typedef typename nodebucket::insertpos insertpos;
    typedef typename hcore::hashpos hashpos;

private:
    hcore m_core;

public:
    class iterator
    {
    public:
        hashpos m_pos;
        iterator() {}
        iterator(hashpos pos) : m_pos(pos) {}
        iterator(const iterator &rhs) { m_pos = rhs.m_pos; }
        iterator &operator=(const iterator &rhs) { m_pos = rhs.m_pos; return *this; }
        bool operator==(const iterator &rhs) const { if( m_pos == rhs.m_pos ) return true; return false; }
        bool operator!=(const iterator &rhs) const { return ! (*this == rhs); }
        iterator operator++() { return iterator(++m_pos); }
        iterator operator++(int)  { iterator it = *this; ++(*this); return it; }
        iterator operator--() { return iterator(--m_pos); }
        iterator operator--(int) { iterator it = *this; --(*this); return it; }
        kvpair &operator*() const { return *m_pos.m_bucket->getptr(m_pos.m_idx); }
        kvpair *operator->() const { return m_pos.m_bucket->getptr(m_pos.m_idx); }
    };

    class const_iterator
    {
    public:
        hashpos m_pos;
        const_iterator() {}
        const_iterator(hashpos pos) : m_pos(pos) {}
        const_iterator(const const_iterator &rhs) { m_pos = rhs.m_pos; }
        const_iterator &operator=(const const_iterator &rhs) { m_pos = rhs.m_pos; return *this; }
        bool operator==(const const_iterator &rhs) const { if( m_pos == rhs.m_pos ) return true; return false; }
        bool operator!=(const const_iterator &rhs) const { return ! (*this == rhs); }
        const_iterator operator++() { return const_iterator(++m_pos); }
        const_iterator operator++(int)  { const_iterator it = *this; ++(*this); return it; }
        const_iterator operator--() { return const_iterator(--m_pos); }
        const_iterator operator--(int) { const_iterator it = *this; --(*this); return it; }
        const kvpair &operator*() const { return *m_pos.m_bucket->getptr(m_pos.m_idx); }
        const kvpair *operator->() const { return m_pos.m_bucket->getptr(m_pos.m_idx); }
    };

    hashmap() {}
    hashmap(const hashmap &rhs) : m_core(rhs.m_core) {}

    hashmap &operator=(const hashmap &rhs)
    {
        m_core = rhs.m_core;
        return *this;
    }

    iterator begin()
    {
        return iterator(m_core.begin());
    }

    iterator end()
    {
        return iterator(m_core.end());
    }

    const_iterator begin() const
    {
        return const_iterator(m_core.begin());
    }
    const_iterator end() const
    {
        return const_iterator(m_core.end());
    }

    bool empty() const { return m_core.empty(); }
    size_t size() const { return m_core.size(); }

    // Mostly useless, except for testing that split/merge happened properly.
    int bucketcount() const { return m_core.bucketcount(); }

    Value &operator[](const Key &key)
    {
        kvpair kv(key, Value());
        pair<hashpos,bool> pos = m_core.insert(kv);
        return pos.first.m_bucket->getptr(pos.first.m_idx)->second;
    }
    
    pair<iterator,bool> insert(const Key &key, const Value &value)
    {
        kvpair kv(key, value);
        pair<hashpos,bool> pos = m_core.insert(kv);
        return pair<iterator,bool>(iterator(pos.first),pos.second);
    }

    void erase(iterator iter) { m_core.erase(hashpos(iter.m_pos)); }

    bool erase(const Key &key)
    {
        kvpair kv(key, Value());
        return m_core.erase(kv);
    }

    void clear() { m_core.clear(); }

    iterator find(const Key &t)
    {
        kvpair kv(t, Value());
        hashpos pos = m_core.find(kv);
        return iterator(pos);
    }

    const_iterator find(const Key &t) const
    {
        kvpair kv(t, Value());
        hashpos pos = m_core.find(kv);
        return const_iterator(pos);
    }

    bool operator==(const hashmap &rhs) const { return m_core.operator==(rhs.m_core); }
    bool operator!=(const hashmap &rhs) const { return ! m_core.operator==(rhs.m_core); }
};

#endif //__hashmap_h


