// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __intset_h
#define __intset_h

#include "glrexport.h"
#include "vec.h"

class LIBGLR_EXPORT IntSetRange
{
public:
    int m_min, m_max;
    IntSetRange();
    IntSetRange(unsigned int min, unsigned int max);
    bool operator==(const IntSetRange &rhs) const;
    bool operator!=(const IntSetRange &rhs) const;
    bool operator&(const IntSetRange &rhs) const;          // test for intersection.
    size_t count() const;
};

class LIBGLR_EXPORT UIntSetRange
{
public:
    unsigned int m_min, m_max;
    UIntSetRange();
    UIntSetRange(unsigned int min, unsigned int max);
    bool operator==(const UIntSetRange &rhs) const;
    bool operator<(const UIntSetRange &rhs) const;
    bool leftof(const UIntSetRange &rhs) const;
    bool operator!=(const UIntSetRange &rhs) const;
    bool operator&(const UIntSetRange &rhs) const;          // test for intersection.
    UIntSetRange intersect(const UIntSetRange &rhs) const;
    unsigned int count() const;
	bool is_subset_of(const UIntSetRange &rhs) const;
	bool contiguous_with(const UIntSetRange &rhs) const;
	UIntSetRange contiguous_range(const UIntSetRange &rhs) const;
};

class LIBGLR_EXPORT UIntSet
{
protected:
    class IntSetPacket
    {
    private:
        int m_ref;
    public:
        IntSetPacket();
        IntSetPacket(const IntSetPacket &rhs);
        IntSetPacket &operator=(const IntSetPacket &rhs);
        int IncRef();
        int DecRef();
        int Refs();
		vector<UIntSetRange> m_ranges;
    };
    IntSetPacket *m_packet;
    void PrepareForInplaceUpdate();

public:
	class UIntSetIterator
	{
	protected:
		UIntSet *m_pIntSet;
		vector<UIntSetRange>::iterator m_iter;
		bool m_bEnd;
	public:
        UIntSetIterator() : m_pIntSet(0), m_iter(0), m_bEnd(true) {}

		UIntSetIterator(UIntSet *pIntSet, vector<UIntSetRange>::iterator iter, bool bEnd)
			: m_pIntSet(pIntSet), m_iter(iter), m_bEnd(bEnd)
		{
		}

		UIntSetRange &operator*()
		{
			return *m_iter;
		}

		UIntSetRange *operator->()
		{
			return m_iter;
		}

		bool operator==(const UIntSetIterator &rhs) const
		{
			if( m_pIntSet != rhs.m_pIntSet )
				return false;
			if( m_bEnd && rhs.m_bEnd )
				return true;
			return m_iter == rhs.m_iter;
		}

		bool operator!=(const UIntSetIterator &rhs) const
		{
			return ! ((*this) == rhs);
		}

		UIntSetIterator& operator++()
		{
			if( m_bEnd )
				return *this;
			++m_iter;
			return *this;
		}

		UIntSetIterator operator++(int)
		{
			if( m_bEnd )
				return UIntSetIterator(m_pIntSet, m_iter, m_bEnd);
			return UIntSetIterator(m_pIntSet, m_iter++, m_bEnd);
		}
	};

    class UIntSetConstIterator
	{
	protected:
		const UIntSet *m_pIntSet;
		vector<UIntSetRange>::const_iterator m_iter;
		bool m_bEnd;
	public:
		UIntSetConstIterator(const UIntSet *pIntSet, vector<UIntSetRange>::const_iterator iter, bool bEnd)
			: m_pIntSet(pIntSet), m_iter(iter), m_bEnd(bEnd)
		{
		}

		const UIntSetRange &operator*()
		{
			return *m_iter;
		}

		const UIntSetRange *operator->()
		{
			return m_iter;
		}

		bool operator==(const UIntSetConstIterator &rhs) const
		{
			if( m_pIntSet != rhs.m_pIntSet )
				return false;
			if( m_bEnd && rhs.m_bEnd )
				return true;
			return m_iter == rhs.m_iter;
		}

		bool operator!=(const UIntSetConstIterator &rhs) const
		{
			return ! ((*this) == rhs);
		}

		UIntSetConstIterator& operator++()
		{
			if( m_bEnd )
				return *this;
			++m_iter;
			return *this;
		}

		UIntSetConstIterator operator++(int)
		{
			if( m_bEnd )
				return UIntSetConstIterator(m_pIntSet, m_iter, m_bEnd);
			return UIntSetConstIterator(m_pIntSet, m_iter++, m_bEnd);
		}
	};

	typedef UIntSetIterator iterator;
	typedef UIntSetConstIterator const_iterator;
    UIntSet();
    UIntSet(const UIntSet &rhs);
    ~UIntSet();
    UIntSet &operator=(const UIntSet &rhs);
    bool operator==(const UIntSet &rhs) const;
    bool operator<(const UIntSet &rhs) const;
    bool operator!=(const UIntSet &rhs) const;
    void operator+=(unsigned int value);
    void operator+=(const UIntSet &rhs);
    void operator+=(const UIntSetRange &rhs);
    void operator-=(unsigned int value);
    void operator-=(const UIntSet &rhs);
    void operator-=(const UIntSetRange &rhs);
    UIntSet &operator%(unsigned int value);
    UIntSet &operator%(const UIntSet &rhs);
    UIntSet &operator%(const UIntSetRange &rhs);
    UIntSet operator~()  const;
    bool operator&(unsigned int value) const;
    size_t count() const;
    size_t paircount() const;
    UIntSet operator&(const UIntSet &rhs) const;
    void operator&=(const UIntSet &rhs);
	iterator begin();
    iterator end();
	const_iterator begin() const;
    const_iterator end() const;
    iterator findintersecting(unsigned int low, unsigned int high);
    void clear();
    void erase();
	bool TestValues(unsigned int *values, size_t valuecount) const;
	bool SanityCheck() const;
};

template<>
class simplehash <UIntSet>
{
public:
    uint32_t operator()(const UIntSet &v) const
    {
        if( ! v.paircount() )
            return 0;
        vector<unsigned int> values;
        for( UIntSet::const_iterator cur = v.begin(), end = v.end(); cur != end; ++cur )
        {
            values.push_back(cur->m_min);
            values.push_back(cur->m_max);
        }
        return hash_uint32s(values.begin(), values.size(), 101);
    }
};

#endif // __intset_h

