// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "rx.h"
#include <ctype.h>
#include "nfa.h"
#include "chtest.h"
#include "chsets.h"

UIntSet LookupNamedCharSet(const char *name, bool *pbFound)
{
    if( strcmp(name,"[:alpha:]") == 0 )
        return named_charset_alpha();
    if( strcmp(name,"[:lower:]") == 0 )
        return named_charset_lower();
    if( strcmp(name,"[:upper:]") == 0 )
        return named_charset_upper();
    if( strcmp(name,"[:alnum:]") == 0 )
        return named_charset_alnum();
    if( strcmp(name,"[:blank:]") == 0 )
        return named_charset_blank();
    if( strcmp(name,"[:space:]") == 0 )
        return named_charset_space();
    if( strcmp(name,"[:digit:]") == 0 )
        return named_charset_digit();
    if( strcmp(name,"[:xdigit:]") == 0 )
        return named_charset_xdigit();
    return UIntSet();
}

Rx::Rx()
{
}

Rx::~Rx()
{
}

RxConcat::RxConcat(Rx *lhs, Rx *rhs)
: m_lhs(lhs), m_rhs(rhs)
{
}

RxConcat::~RxConcat()
{
    delete m_lhs;
    delete m_rhs;
}

Rx *RxConcat::Copy()
{
    return new RxConcat(m_lhs->Copy(), m_rhs->Copy());
};

void RxConcat::CaseDesensitize()
{
    m_lhs->CaseDesensitize();
    m_rhs->CaseDesensitize();
}

int RxConcat::AddToNfa(Nfa &nfa, int startstate)
{
    int finalstate = m_lhs->AddToNfa(nfa, startstate);
    finalstate = m_rhs->AddToNfa(nfa, finalstate);
    return finalstate;
}

bool RxConcat::EqualTo(Rx *rhs)
{
    RxConcat *prhs = dynamic_cast<RxConcat*>(rhs);
    if( ! prhs )
        return false;
    return m_lhs->EqualTo(prhs->m_lhs) && m_rhs->EqualTo(prhs->m_rhs);
}

RxUnion::RxUnion(Rx *lhs, Rx *rhs) : m_lhs(lhs), m_rhs(rhs)
{
}

RxUnion::~RxUnion()
{
    delete m_lhs;
    delete m_rhs;
}

Rx *RxUnion::Copy()
{
    return new RxUnion(m_lhs->Copy(), m_rhs->Copy());
}

void RxUnion::CaseDesensitize()
{
    m_lhs->CaseDesensitize();
    m_rhs->CaseDesensitize();
}

int RxUnion::AddToNfa(Nfa &nfa, int startstate)
{
    int finalstate = m_lhs->AddToNfa(nfa, startstate);
    int finalstate2 = m_rhs->AddToNfa(nfa, startstate);
    int finalstate3 = nfa.addstate();
    nfa.addtransition_empty(finalstate, finalstate3);
    nfa.addtransition_empty(finalstate2, finalstate3);
    return finalstate3;
}

bool RxUnion::EqualTo(Rx *rhs)
{
    RxUnion *prhs = dynamic_cast<RxUnion*>(rhs);
    if( ! prhs )
        return false;
    return m_lhs->EqualTo(prhs->m_lhs) && m_rhs->EqualTo(prhs->m_rhs);
}

RxStar::RxStar(Rx *rx) : m_rx(rx)
{
}

RxStar::~RxStar()
{
    delete m_rx;
}

Rx *RxStar::Copy()
{
    return new RxStar(m_rx->Copy());
}

void RxStar::CaseDesensitize()
{
    m_rx->CaseDesensitize();
}

int RxStar::AddToNfa(Nfa &nfa, int startstate)
{
    int secondstate = nfa.addstate();
    int finalstate = m_rx->AddToNfa(nfa, secondstate);
    nfa.relocatedst(finalstate, secondstate);
    nfa.addtransition_empty(startstate, secondstate);
    nfa.addtransition_empty(secondstate, finalstate);
    return finalstate;
}

bool RxStar::EqualTo(Rx *rhs)
{
    RxStar *prhs = dynamic_cast<RxStar*>(rhs);
    if( ! prhs )
        return false;
    return m_rx->EqualTo(prhs->m_rx);
}

RxPlus::RxPlus(Rx *rx) : m_rx(rx)
{
}

RxPlus::~RxPlus()
{
    delete m_rx;
}

Rx *RxPlus::Copy()
{
    return new RxPlus(m_rx->Copy());
}

void RxPlus::CaseDesensitize()
{
    m_rx->CaseDesensitize();
}

int RxPlus::AddToNfa(Nfa &nfa, int startstate)
{
    int secondstate = nfa.addstate();
    nfa.addtransition_empty(startstate, secondstate);
    int finalstate = m_rx->AddToNfa(nfa, secondstate);
    nfa.addtransition_empty(finalstate, secondstate);
    finalstate = m_rx->AddToNfa(nfa, secondstate);
    return finalstate;
}

bool RxPlus::EqualTo(Rx *rhs)
{
    RxPlus *prhs = dynamic_cast<RxPlus*>(rhs);
    if( ! prhs )
        return false;
    return m_rx->EqualTo(prhs->m_rx);
}

RxOpt::RxOpt(Rx *rx) : m_rx(rx)
{
}

RxOpt::~RxOpt()
{
    delete m_rx;
}

Rx *RxOpt::Copy()
{
    return new RxOpt(m_rx->Copy());
}

void RxOpt::CaseDesensitize()
{
    m_rx->CaseDesensitize();
}

int RxOpt::AddToNfa(Nfa &nfa, int startstate)
{
    int finalstate = m_rx->AddToNfa(nfa, startstate);
    nfa.addtransition_empty(startstate, finalstate);
    return finalstate;
}

bool RxOpt::EqualTo(Rx *rhs)
{
    RxOpt *prhs = dynamic_cast<RxOpt*>(rhs);
    if( ! prhs )
        return false;
    return m_rx->EqualTo(prhs->m_rx);
}

RxMany::RxMany(Rx *rx, int min, int max)
: m_rx(rx), m_min(min), m_max(max)
{
}

RxMany::~RxMany()
{
    delete m_rx;
}

Rx *RxMany::Copy()
{
    return new RxMany(m_rx->Copy(), m_min, m_max);
}

void RxMany::CaseDesensitize()
{
    m_rx->CaseDesensitize();
}

int RxMany::AddToNfa(Nfa &nfa, int startstate)
{
    int finalstate = startstate;
    struct range { int count; int min; int max; } rng = {-1, -1, -1};
    rng.min = m_min;
    rng.max = m_max;
    if( rng.min == rng.max )
        rng.count = rng.min;
    if( rng.count != -1 ) // Fixed count
    {
        // Fixed count, treat like concatination.
        for( int i = 0, m = rng.count; i < m; ++i )
            finalstate = m_rx->AddToNfa(nfa, finalstate);
    }
    else if( rng.min != -1 && rng.max != -1 ) // Range
    {
        for( int i = 0, m = rng.min; i < m; ++i )
            finalstate = m_rx->AddToNfa(nfa, finalstate);
        vector<int> intermediates;
        intermediates.reserve(rng.max-rng.min);
        for( int i = 0, m = rng.max-rng.min; i < m; ++i )
        {
            intermediates.append(finalstate);
            finalstate = m_rx->AddToNfa(nfa, finalstate);
        }
        for( size_t i = 0, m = intermediates.size(); i < m; ++i )
            nfa.addtransition_empty(intermediates[i], finalstate);
    }
    else if( rng.min == -1 && rng.max != -1 ) // No min
    {
        // No min, treat like concatination with empty transitions on each intermediate finish.
        vector<int> intermediates;
        intermediates.reserve(rng.max);
        for( int i = 0, m = rng.max; i < m; ++i )
        {
            intermediates.append(finalstate);
            finalstate = m_rx->AddToNfa(nfa, finalstate);
        }
        for( int i = 0, m = rng.max; i < m; ++i )
            nfa.addtransition_empty(intermediates[i], finalstate);
    }
    else if( rng.min != -1 && rng.max == -1 ) // No max.
    {
        // No max, treat like concatination with empty transition to first to last finish.
        for( int i = 0, m = rng.min-1; i < m; ++i )
            finalstate = m_rx->AddToNfa(nfa, finalstate);
        int nexttolast = finalstate;
        finalstate = m_rx->AddToNfa(nfa, finalstate);
        nfa.addtransition_empty(finalstate, nexttolast);
        nexttolast = finalstate;
        finalstate = nfa.addstate();
        nfa.addtransition_empty(nexttolast, finalstate);
    }
    else
    {
        // Same as '*'
        finalstate = m_rx->AddToNfa(nfa, startstate);
        nfa.relocatedst(finalstate, startstate);
        nfa.addtransition_empty(startstate, finalstate);
    }
    return finalstate;
}

bool RxMany::EqualTo(Rx *rhs)
{
    RxMany *prhs = dynamic_cast<RxMany*>(rhs);
    if( ! prhs )
        return false;
    if( m_min != prhs->m_min || m_max != prhs->m_max )
        return false;
    return m_rx->EqualTo(prhs->m_rx);
}

RxChars::RxChars()
{
}

RxChars::RxChars(int c)
{
    m_chars += c;
}

RxChars::RxChars(const UIntSet &chars) : m_chars(chars)
{
}

RxChars::~RxChars()
{
}

Rx *RxChars::Copy()
{
    return new RxChars(m_chars);
}

const UIntSet &named_charset_lower();
const UIntSet &named_charset_upper();

void RxChars::CaseDesensitize()
{
	// Fast on even massive ranges.
	UIntSet lowers = m_chars & named_charset_lower();
	UIntSet uppers = m_chars & named_charset_upper();
    for( UIntSet::iterator a = lowers.begin(); a != lowers.end(); ++a )
    {
        for( unsigned int i = a->m_min, endi = a->m_max; i <= endi; ++i )
        {
            unsigned int highi = glr_toupper(i);
			if( i != highi )
				m_chars += highi;
        }
    }
    for( UIntSet::iterator a = uppers.begin(); a != uppers.end(); ++a )
    {
        for( unsigned int i = a->m_min, endi = a->m_max; i <= endi; ++i )
        {
            unsigned int lowi = glr_tolower(i);
			if( i != lowi )
				m_chars += lowi;
        }
    }
}

int RxChars::AddToNfa(Nfa &nfa, int startstate)
{
    int finalstate = nfa.addstate();
    nfa.addtransition_charset(startstate, finalstate, m_chars);
    return finalstate;
}

bool RxChars::EqualTo(Rx *rhs)
{
    RxChars *prhs = dynamic_cast<RxChars*>(rhs);
    if( ! prhs )
        return false;
    return (m_chars == prhs->m_chars);
}

RxCapture::RxCapture(Rx *rx) : m_rx(rx)
{
}

RxCapture::~RxCapture()
{
    delete m_rx;
}

Rx *RxCapture::Copy()
{
    return new RxCapture(m_rx->Copy());
}

void RxCapture::CaseDesensitize()
{
    m_rx->CaseDesensitize();
}

int RxCapture::AddToNfa(Nfa &nfa, int startstate)
{
    int a = nfa.addstate();
    nfa.markstate(a, STATE_MARKER_STARTCAPTURE);
    nfa.addtransition_empty(startstate, a);
    int b = m_rx->AddToNfa(nfa, a);
    int finalstate = nfa.addstate();
    nfa.markstate(finalstate, STATE_MARKER_ENDCAPTURE);
    nfa.addtransition_empty(b, finalstate);
    return finalstate;
}

bool RxCapture::EqualTo(Rx *rhs)
{
    RxCapture *prhs = dynamic_cast<RxCapture*>(rhs);
    if( ! prhs )
        return false;
    return m_rx->EqualTo(prhs->m_rx);
}

Rx *RxApplyOpenClose(Rx *rx, int rxopen, int rxclose)
{
    if( rxopen & RX_CASEI )
        rx->CaseDesensitize();
    return rx;
}
