// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "sysport.h"
#ifdef _WIN32
#include <windows.h>
#include <winnt.h>
#else
#endif

#ifdef _WIN32

static void ptd_thread_register();
static void ptd_thread_free();

BOOL WINAPI DllMain(
  __in  HINSTANCE hinstDLL,
  __in  DWORD fdwReason,
  __in  LPVOID lpvReserved
)
{
    switch(fdwReason)
    {
    case DLL_PROCESS_ATTACH:
        ptd_thread_register();
        break;
    case DLL_THREAD_ATTACH:
        ptd_thread_register();
        break;
    case DLL_THREAD_DETACH:
        ptd_thread_free();
        break;
    case DLL_PROCESS_DETACH:
        break;
    default:
        break;
    }
    return TRUE;
}

typedef struct _TEB {
  BYTE  Reserved1[1952];
  PVOID Reserved2[412];
  PVOID TlsSlots[64];
  BYTE  Reserved3[8];
  PVOID Reserved4[26];
  PVOID ReservedForOle;
  PVOID Reserved5[4];
  VOID **TlsExpansionSlots;
} TEB, *PTEB;

#define MAX_PTD (1088)

// Memory page definition
struct tag_mem_page;
typedef struct tag_mem_page mem_page;
struct tag_mem_page
{
    mem_page *next;
    uint32_t pagesize;
    uint32_t capacity;
    uint32_t count;
    uint32_t objsize;
};

// PTD Release pointers and spinlock
static uint32_t g_release_set[MAX_PTD/32] = {   0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                                0, 0, 0, 0};
typedef void (*releasefnct)(void*);
static releasefnct g_release_ptrs[MAX_PTD];
static spin_t g_tls_slock = 0;

// List of running threads and their TEB data (for x-thread releasing of PTD data)
struct tag_threadchain;
typedef struct tag_threadchain threadchain;
struct tag_threadchain
{
    TEB *teb;
    threadchain *prev;
    threadchain *next;
};
static threadchain *g_threadroot = 0, *g_threadtail = 0;
static mem_page* g_tls_root = 0;

// System page size cache
static size_t g_pagesize = 0;

// CRITICAL_SECTION allocation and spinlock
static spin_t g_csalloc_slock = 0;
static mem_page* g_csec_root = 0;

static void *GetSlot(TEB *teb, ptd_t ptd)
{
    if( ptd > MAX_PTD )
        return 0;
    if( ptd < 64 )
        return teb->TlsSlots[ptd];
    return teb->TlsExpansionSlots[ptd-64];
}

static mem_page *page_create(uint32_t objsize)
{
    mem_page *page = 0;
    uint32_t *flags = 0;
    int i, flagcount;
    ptrdiff_t pagesize = 0;
    for(i=0;i<32;++i)
    {
        if( (objsize>>i) == 1 )
        {
            if( objsize > ((uint32_t)1<<i) )
                objsize = ((uint32_t)1<<(i+1));
            else
                objsize = ((uint32_t)1<<i);
            break;
        }
    }
    page = (mem_page*)sys_malloc(1, &pagesize, 0);
    page->pagesize = pagesize;
    page->capacity = (((pagesize-sizeof(mem_page))*8/(1+objsize*8))>>5)<<5;
    page->count = 0;
    page->next = 0;
    page->objsize = objsize;
    flags = (uint32_t*)(page+1);
    for(i=0,flagcount = page->capacity>>5;i<flagcount;++i)
        flags[i] = 0;
    return page;
}

static void *page_alloc_data(mem_page *page)
{
    uint32_t *flags = 0, flag = 0;
    int index, flagidx = 0, bitidx = 0, flagcount;
    void *ret = 0;
    if( page->count == page->capacity )
        return 0;
    flags = (uint32_t*)(page+1);
    for(flagcount=page->capacity>>5;flagidx<flagcount;++flagidx)
    {
        if( ~flags[flagidx] == 0 )
            continue;
        flag = flags[flagidx];
        while( flag & 0x01 )
        {
            ++bitidx;
            flag >>= 1;
        }
        flags[flagidx] |= 1<<bitidx;
        index = flagidx*32+bitidx;
        ret = ((char*)(flags+flagcount))+(page->objsize*index);
        ++(page->count);
        break;
    }
    return ret;
}

static void page_free_data(void *vp)
{
    uint32_t *flags = 0;
    mem_page *page = 0;
    int flagcount = 0;
    int index, flagidx, bitidx;

    page = (mem_page*)((((ptrdiff_t)vp)/g_pagesize)*g_pagesize);
    flagcount = page->capacity>>5;
    flags = (uint32_t*)(page+1);
    index = ((char*)vp-(char*)(flags+flagcount))/page->objsize;
    flagidx = index>>5;
    bitidx = index&0x001f;
    flags[flagidx] &= ~(1<<bitidx);
    (page->count)--;
}

static void *chain_alloc_data(mem_page **pppage, uint32_t objsize)
{
    mem_page *page = *pppage, *tmp = 0;
    void *vpdata = 0;
    while( page )
    {
        vpdata = page_alloc_data(page);
        if( vpdata )
            break;
        page = page->next;
    }
    if( ! page )
    {
        page = page_create(objsize);
        page->next = *pppage;
        *pppage = page;
        vpdata = page_alloc_data(page);
    }
    return vpdata;
}

static void ptd_thread_register()
{
    int i = 0;
    threadchain *tc = 0;
    TEB *teb = NtCurrentTeb();
    mem_page *curpage = 0;

    SPIN(g_tls_slock,i,100);
    curpage = g_tls_root;
    tc = chain_alloc_data(&g_tls_root, sizeof(threadchain));
    tc->teb = teb;
    tc->next = 0;
    tc->prev = g_threadtail;
    if( g_threadtail )
        g_threadtail->next = tc;
    g_threadtail = tc;
    if( ! g_threadroot )
        g_threadroot = tc;
    UNSPIN(g_tls_slock);
}

static void ptd_thread_free()
{
    int i = 0, ptdidx = 0, ptdflagidx = 0, ptdbitidx = 0;
    void (*release)(void*) = 0;
    void *vpdata = 0;
    threadchain *tc = 0;
    TEB *teb = NtCurrentTeb();
    uint32_t flag = 0;

    SPIN(g_tls_slock,i,100);
    // remove from chain, if tracked
    tc = g_threadroot;
    while( tc )
    {
        if( tc->teb != teb )
        {
            tc = tc->next;
            continue;
        }
        if( tc->prev )
            tc->prev->next = tc->next;
        if( tc->next )
            tc->next->prev = tc->prev;
        if( tc == g_threadroot )
            g_threadroot = tc->next;
        if( tc == g_threadtail )
            g_threadtail = tc->prev;
        page_free_data(tc);
        break;
    }
    for(ptdflagidx=0;ptdflagidx<MAX_PTD/32;++ptdflagidx)
    {
        flag = g_release_set[ptdflagidx];
        if( ! flag )
            continue;
        for(ptdbitidx=0;ptdbitidx<32;++ptdbitidx,flag>>=1)
        {
            if( ! (flag & 1) )
                continue;
            ptdidx = (ptdflagidx<<5)+ptdbitidx;
            release = g_release_ptrs[ptdidx];
            if( ! release )
                continue;
            vpdata = GetSlot(teb,ptdidx);
            if( vpdata )
                release(vpdata);
        }
    }
    UNSPIN(g_tls_slock);
}

typedef struct
{
    thread_fnct f;
    void *vpdata;
    spin_t ready;
} tpdata;

static DWORD WINAPI tpThreadProc(void *lpParameter)
{
    tpdata *ptp = (tpdata*)lpParameter;
    thread_fnct f = ptp->f;
    void *vpdata = ptp->vpdata;
    InterlockedExchange(&ptp->ready, 1);
    return f(vpdata);
}

static size_t GetSystemPageSize()
{
    SYSTEM_INFO si;
    GetSystemInfo(&si); 
    return si.dwPageSize;
}
#endif

void *sys_malloc(ptrsize_t size, ptrsize_t *osize, int iflags)
{
#ifdef _WIN32
    if( ! size )
        return 0;
    if( ! g_pagesize )
        g_pagesize = GetSystemPageSize();
    size = (size + (g_pagesize-1)) & ~(g_pagesize-1);
    if( osize )
        *osize = size;
    return VirtualAlloc(0, size, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
#else
#endif
}

int sys_free(void *vpdata)
{
#ifdef _WIN32
    return VirtualFree(vpdata, 0, MEM_RELEASE) ? 1 : 0;
#else
#endif
}

threadid_t start_thread(thread_fnct f, void *vpdata)
{
#ifdef _WIN32
    DWORD threadid;
    int i;
    tpdata tp = {f, vpdata, 0};
    if( ! CreateThread(NULL, 0, tpThreadProc, (void*)&tp, 0, &threadid) )
        return -1;
    while( InterlockedCompareExchange(&tp.ready, 0, 0) == 0 )
        for(i=0;i<100;++i);
    return threadid;
#else
#endif
}

unsigned int current_threadid()
{
#ifdef _WIN32
    return (unsigned int)GetCurrentThreadId();
#else
#endif
}

ptd_t ptd_alloc(void (*release)(void*))
{
#ifdef _WIN32
    int i = 0;
    ptd_t ret = TlsAlloc();
    SPIN(g_tls_slock,i,100);
    g_release_set[ret/32] |= (1<<(ret%32));
    g_release_ptrs[ret] = release;
    UNSPIN(g_tls_slock);
    return ret;
#endif
}

void ptd_free(ptd_t ptd)
{
#ifdef _WIN32
    int i = 0;
    threadchain *tc = 0;
    void (*release)(void*) = 0;
    void *vpdata = 0;
    if( ptd > MAX_PTD )
        return;
    SPIN(g_tls_slock,i,100);
    if( (g_release_set[ptd/32] & 1<<(ptd%32)) )
    {
        release = g_release_ptrs[ptd];
        if( release )
        {
            tc = g_threadroot;
            while( tc )
            {
                vpdata = GetSlot(tc->teb,ptd);
                if( vpdata )
                    release(vpdata);
                tc = tc->next;
            }
        }
        g_release_set[ptd/32] &= ~(1<<(ptd%32));
    }
    UNSPIN(g_tls_slock);
    TlsFree(ptd);
#else
#endif
}

void *ptd_get(ptd_t ptd)
{
#ifdef _WIN32
	return TlsGetValue(ptd);
#else
#endif
}

void ptd_set(ptd_t ptd, void *vp)
{
#ifdef _WIN32
	TlsSetValue(ptd, vp);
#else
#endif
}
