// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "tokensetmerge.h"
#include "sortedset.h"

ProductionInState::ProductionInState()
: m_p(0), m_state(0), m_clone(0), m_altered_production(0)
{
}

ProductionInState::ProductionInState(Production *p, OrderedTokenSet *state)
: m_p(p), m_state(state), m_clone(0), m_altered_production(0)
{
}

bool ProductionInState::operator<(const ProductionInState &rhs) const
{
    if( m_p < rhs.m_p )
        return true;
    else if( m_p == rhs.m_p && m_state < rhs.m_state )
        return true;
    return false;
}

bool ProductionInState::operator==(const ProductionInState &rhs) const
{
    return (m_p == rhs.m_p) && (m_state == rhs.m_state);
}

SymbolInState::SymbolInState()
: m_s(0), m_state(0), m_clone(0), m_aggregate_nt_actions(0)
{
}

SymbolInState::SymbolInState(Symbol *s, OrderedTokenSet *state)
: m_s(s), m_state(state), m_clone(0), m_aggregate_nt_actions(0)
{
}

bool SymbolInState::operator<(const SymbolInState &rhs) const
{
    if( m_s < rhs.m_s )
        return true;
    else if( m_s == rhs.m_s && m_state < rhs.m_state )
        return true;
    return false;
}

bool SymbolInState::operator==(const SymbolInState &rhs) const
{
    return (m_s == rhs.m_s) && (m_state == rhs.m_state);
}

AlteredProductionInState::AlteredProductionInState(Production *p, OrderedTokenSet *state)
: m_p(p), m_state(state)
{
    m_items.reserve(p->m_items.size());
    for( vector<Symbol*>::iterator cursymbol = p->m_items.begin(); cursymbol != p->m_items.end(); ++cursymbol )
        m_items.push_back( SymbolInState(*cursymbol, state) );
}

TokenSetMergeAlgorithm::TokenSetMergeAlgorithm(Grammar *g, outbuf *err) : m_g(g), m_err(err)
{
}

TokenSetMergeAlgorithm::~TokenSetMergeAlgorithm()
{
}

bool TokenSetMergeAlgorithm::IsSymbolCloneable(Symbol *s, OrderedTokenSet *tokset, bool bForceYes)
{
    SType stype = s->GetSType();
    switch( stype )
    {
    case CHARSET_SYMBOL:
        return true;
    case NONTERMINAL_SYMBOL:
        break;
    default:
        break;
    };
    vector<SymbolInState>::iterator iter = m_clones.find( SymbolInState(s,tokset) );
    if( iter != m_clones.end() )
        return true;
    if( bForceYes || stype == NONTERMINAL_SYMBOL )
    {
        m_clones.push_back( SymbolInState(s,tokset) );
        return true;
    }
    return false;
}

void TokenSetMergeAlgorithm::FinalizeActions(Symbol *s, OrderedTokenSet *tokset, TokenSetActions &finalactions)
{
    vector<SymbolInState>::iterator iter = m_clones.find(SymbolInState(s,tokset));
    if( iter == m_clones.end() )
    {
        if( m_err )
            ob_puts(m_err, "error: Internal state inconsistant when finalizing token actions.");
        m_g->m_errcount++;
        throw MergeTokenSetStatesException();
    }
    if( ! iter->m_aggregate_nt_actions )
    {
        TokenSetActions *pfinalactions = new TokenSetActions;
        *pfinalactions = finalactions;
        iter->m_aggregate_nt_actions = pfinalactions;
    }
    else if( *(iter->m_aggregate_nt_actions) != finalactions )
    {
        m_g->m_wrncount++;
        if( m_err )
        {
            ob_puts(m_err, "warning: Token state stack inconsistant for nonterminal ");
            s->Print(m_err);
            ob_putc(m_err, '\n');
            finalactions.Print(m_err);
            ob_puts(m_err, "Based on previous productions of ");
            s->Print(m_err);
            ob_puts(m_err, " different actions were expected:\n");
            iter->m_aggregate_nt_actions->Print(m_err);
        }
        throw MergeTokenSetStatesException();
    }
}

bool TokenSetMergeAlgorithm::GetFinalActions(Symbol *s, OrderedTokenSet *tokset, TokenSetActions* &aggregate_nt_actions)
{
    vector<SymbolInState>::iterator iter = m_clones.find(SymbolInState(s,tokset));
    if( iter == m_clones.end() || iter->m_aggregate_nt_actions == 0 )
        return false;
    aggregate_nt_actions = iter->m_aggregate_nt_actions;
    return true;
}

AlteredProductionInState *TokenSetMergeAlgorithm::GetProductionInState(Production *p, OrderedTokenSet *tokset)
{
    vector<ProductionInState>::iterator iter = m_productions.find( ProductionInState(p,tokset) );
    if( iter != m_productions.end() && iter->m_altered_production )
        return iter->m_altered_production;
    AlteredProductionInState *ret = new AlteredProductionInState(p,tokset);
    if( iter == m_productions.end() )
    {
        ProductionInState pinstate(p,tokset);
        pinstate.m_altered_production = ret;
        m_productions.push_back(pinstate);
    }
    else
        iter->m_altered_production = ret;
    return ret;
}

void TokenSetMergeAlgorithm::RecordTokenSetMiss(SymbolInState s, ProductionInState p, int pos)
{
	m_TokenSetMiss.insert( TokenSetMiss(s, p, pos) );
}

static vector<Production*> *g_vec;

static int SortByOtherVector(Production *const *lhs, Production *const *rhs)
{
    Production *const plhs = *lhs;
    Production *const prhs = *rhs;
    int ilhs = g_vec->find(plhs) - g_vec->begin();
    int irhs = g_vec->find(prhs) - g_vec->begin();
    if( ilhs < irhs )
        return -1;
    else if( irhs < ilhs )
        return 1;
    return 0;
}

void TokenSetMergeAlgorithm::Symbol_PreMergeTokenSetStatesRec(Symbol *s, OrderedTokenSet *tokset, vector<OrderedTokenSet*> &statestack, TokenSetActions &finalactions)
{
    if( s->GetSType() == NONTERMINAL_SYMBOL )
    {
        TokenSetActions *aggregate_nt_actions = 0;
        if( ! GetFinalActions(s, tokset, aggregate_nt_actions) )
        {
            // Couldn't find the cached actions for this nonterminal/tokenset, so we'll figure it
            // out by adding the actions for the productions that make up this nonterminal.
            vector<OrderedTokenSet*> resetstack = statestack;
            TokenSetActions localactions;

            // Get the productions, sorted by the m_productionordering vector, which in short means
            // productions which depend on terminals first, followed by productions that depend on
            // the other productions.  This, in combination with the action caching for nonterminals
            // results in an order of production evaluation that never loops back on itself or evaluates
            // a given production twice.
            vector<Production*> owningproductions = s->m_OwningProductions;
            // There is a non thread safe hack here. {
            g_vec = &m_productionordering;
            owningproductions.sort(SortByOtherVector);
            // } non thread safe hack

            // We only need one for the actions, the rest are for consistancy checking (they should
            // all yield the same actions), and so that we spot all terminals that aren't in token sets.
            for( vector<Production*>::iterator curclone = owningproductions.begin(); curclone != owningproductions.end(); ++curclone )
            {
                Production *pcur = *curclone;
                if( pcur->IsErrorProduction() )
                    break; // always placed last, and ignored
                Production_PreMergeTokenSetStatesRec(pcur, tokset, statestack, localactions);
                // Cacheing and consistancy check.
                FinalizeActions(s, tokset, localactions);
                // Reset for the next production.
                statestack = resetstack;
                localactions.m_actions.clear();
            }
        }
        // The actions should be cached at this point.  If not, big problem. 
        if( ! GetFinalActions(s, tokset, aggregate_nt_actions) )
            throw MergeTokenSetStatesException();
	    aggregate_nt_actions->MergeTo(finalactions);
        aggregate_nt_actions->ApplyTo(m_g, statestack, m_err);
        return;
    }

    // All other symbols.
    TokenSetAction *action = tokset ? tokset->FindTokenSetAction(s) : 0;
    if( ! action )
        return; // empty action.
    action->ApplyTo(m_g, statestack, m_err);
    action->MergeTo(finalactions);
}

void TokenSetMergeAlgorithm::Production_PreMergeTokenSetStatesRec(Production *p, OrderedTokenSet *tokset, vector<OrderedTokenSet*> &statestack, TokenSetActions &finalactions)
{
    AlteredProductionInState *realp = GetProductionInState(p,tokset);
    for( vector<SymbolInState>::iterator cursymbolinstate = realp->m_items.begin(); cursymbolinstate != realp->m_items.end(); ++cursymbolinstate )
    {
        Symbol *curs = cursymbolinstate->m_s;
        SymbolInState cursymbolclone(curs, statestack[-1]);
        // If we cannot find the token in the tokenset, use the original non-state associated token, so that we get nondeterministic parsing.
        if( ! IsSymbolCloneable(cursymbolclone.m_s, cursymbolclone.m_state, false) )
        {
            RecordTokenSetMiss(cursymbolclone, ProductionInState(p, realp->m_state), cursymbolinstate-realp->m_items.begin());
            cursymbolclone.m_state = 0;
            if( m_nondeterministic_tokens.find(cursymbolclone) == m_nondeterministic_tokens.end() )
                m_nondeterministic_tokens.push_back(cursymbolclone);
        }
        *cursymbolinstate = cursymbolclone;
        Symbol_PreMergeTokenSetStatesRec(curs, cursymbolclone.m_state, statestack, finalactions);
    }
}

static bool IsCompleteProduction(Production *p, vector<Symbol*> &completenonterminals)
{
    for( vector<Symbol*>::iterator cur = p->m_items.begin(); cur != p->m_items.end(); ++cur )
    {
        if( (*cur)->IsTerminal() )
            continue;
        if( completenonterminals.find(*cur) != completenonterminals.end() )
            continue;
        return false;
    }
    return true;
}

// The goal of this algorithm is to identify tokens that have not been shuffled into one token set
// or another (often an error) and to point out inconsistant nonterminals (always an error), from
// the standpoint that all productions for a given nonterminal must have the same net effect on the
// token set stack, or else the resulting parser will be completely unusable.
//
// To achieve our goal, we want to feed each of the tokens from the productions into a token set
// stack simulator from the states at which we can encounter them.  A complication arises, in that
// we cannot know which states to start with for each production, unless we start at the <Start>
// production and walk down, figuring out the effect of a non-production by recursively evaluating
// the production's net effect.  On the other hand, we must be careful not to loop back over the
// same productions.  To avoid the loop back, we pre-order the productions, putting productions
// that rely on terminals first, then productions that rely on those and so forth.  Then, when we
// encounter a nonterminal during the walk, we walk the productions in the order that we pre-ordered
// all productions in, therefore assuring we don't hit the loop back.
void TokenSetMergeAlgorithm::Collect()
{
    m_TokenSetMiss.clear();

    if( ! m_g->m_TokenSets.size() )
        return;
    if( ! m_g->m_InitialTokenState )
    {
        if( m_err )
            ob_puts(m_err, "warning: Token sets are defined, but no initial TokenSet, skipping token merging and tokenizing steps.\n");
        m_g->m_wrncount++;
        return;
    }

    // Validate that token sets we use are defined.
    for( vector<OrderedTokenSet*>::iterator curTokenSet = m_g->m_TokenSets.begin(); curTokenSet != m_g->m_TokenSets.end(); ++curTokenSet )
    {
        OrderedTokenSet &ots = **curTokenSet;
        for( vector<TokenSetAction>::iterator curaction = ots.beginaction(); curaction != ots.endaction(); ++curaction )
        {
            switch( curaction->m_action )
            {
                case TokenSetAction::GOTO:
                case TokenSetAction::PUSH:
                {
                    if( curaction->m_actionset->m_symbols.size() == 0 )
                    {
                        if( m_err )
                            ob_printf(m_err, "error: %s(%d): The token set <%s> is used, but not defined\n", curaction->m_file.c_str(), curaction->m_lineno, curaction->m_actionset->m_Name.c_str());
                        m_g->m_errcount++;
                    }
                    break;
                }
                default:
                    continue;
            }
        }
    }

    // Mark each token as cloned
    for( vector<OrderedTokenSet*>::iterator curts = m_g->m_TokenSets.begin(); curts != m_g->m_TokenSets.end(); ++curts )
    {
        OrderedTokenSet &tokset = **curts;
        for( vector<Symbol*>::iterator cur = tokset.m_symbols.begin(); cur != tokset.m_symbols.end(); ++cur )
        {
            Symbol *s = *cur;
            IsSymbolCloneable(s, &tokset, true);
        }
    }

    // Make sure there aren't any problems before we continue.    
    if( m_g->m_errcount )
        return;
    
    // Get the total ordering.
    vector<Symbol*> completenonterminals;
    m_productionordering.clear();
    int prevcount = -1;
    while( (int)m_productionordering.size() > prevcount )
    {
        prevcount = m_productionordering.size();
        for( vector<Production*>::iterator curproduction = m_g->m_productions.begin(); curproduction != m_g->m_productions.end(); ++curproduction )
        {
            if( (*curproduction)->IsErrorProduction() )
                continue;
            if( m_productionordering.find(*curproduction) != m_productionordering.end() )
                continue;
            if( IsCompleteProduction(*curproduction, completenonterminals) )
            {
                if( completenonterminals.find((*curproduction)->m_nt) == completenonterminals.end() )
                    completenonterminals.push_back((*curproduction)->m_nt);
                m_productionordering.push_back(*curproduction);
            }
        }
        if( m_productionordering.size() == m_g->m_productions.size() )
            break;
    }
    for( vector<Production*>::iterator curproduction = m_g->m_productions.begin(); curproduction != m_g->m_productions.end(); ++curproduction )
    {
        if( (*curproduction)->IsErrorProduction() )
            m_productionordering.push_back(*curproduction);
    }
    if( m_productionordering.size() < m_g->m_productions.size() )
    {
        if( m_err )
        {
            ob_puts(m_err, "error: Couldn't produce a total token ordering.  Some productions depended on themselves.\n");
            for( vector<Production*>::iterator curproduction = m_g->m_productions.begin(); curproduction != m_g->m_productions.end(); ++curproduction )
            {
                if( m_productionordering.find(*curproduction) != m_productionordering.end() )
                    continue;
                (*curproduction)->Print(m_err);
                ob_puts(m_err, "\n");
            }
        }
        m_g->m_errcount++;
        return;
    }

    // Do the grammar-TokenSet merge algorithm.
    try
    {
        TokenSetActions finalactions;
        vector<OrderedTokenSet*> statestack;
        m_clones.push_back( SymbolInState(m_g->m_startsymbol, m_g->m_InitialTokenState) );
        statestack.append(m_g->m_InitialTokenState);
        Symbol_PreMergeTokenSetStatesRec(m_g->m_startsymbol, m_g->m_InitialTokenState, statestack, finalactions);
    }
    catch( MergeTokenSetStatesException )
    {
        if( m_err )
            ob_puts(m_err, "error: Couldn't Merge TokenSet States\n");
        m_g->m_errcount++;
        return;
    }

    // Make sure there aren't any problems before we continue.    
    if( m_g->m_errcount )
        return;
}

Production *TokenSetMergeAlgorithm::CloneProduction(Production *porig, Symbol *clonent, AlteredProductionInState *altered_production)
{
    Production *pclone = 0;
    if( porig->m_nt == clonent )
        pclone = porig;
    else
    {
        // Adding the production will add it to clonent, so no need to add it ourself.
        pclone = new Production(m_g, clonent, porig->m_items, porig->m_sldecls);
        pclone->m_optionalranges = porig->m_optionalranges;
        pclone->m_copyof = porig;
        porig->m_copycats.append(pclone);
        m_g->m_productions.insertat(porig->GetIndex()+porig->m_copycats.size(), pclone);
        // Don't need to copy forbid rules, ChildConflicts always uses clone parent.
        pclone->m_rank = porig->m_rank;
        pclone->m_reject = porig->m_reject;
        pclone->m_label = porig->m_label;
    }

    // Update each item in the production with the clone for at each item's state.
    if( altered_production )
    {
        for( size_t i = 0, n = altered_production->m_items.size(); i < n; ++i )
        {
            vector<SymbolInState>::iterator iter = m_clones.find(altered_production->m_items[i]);
            if( iter != m_clones.end() )
                pclone->m_items[i] = iter->m_clone;
        }
    }

    return pclone;
}

Symbol* TokenSetMergeAlgorithm::CloneSymbol(Symbol *orig, OrderedTokenSet *ts)
{
    if( orig == m_g->m_startsymbol )
        return orig; // Never clone the start symbol.
    SType stype = orig->GetSType();
    switch(stype)
    {
    case LITERAL_SYMBOL:
        {
        Literal *s = m_g->FindOrAddLiteral(orig->m_Name, orig->IsCaseInsensitive(), ts, true);
        s->m_cloneof = orig;
        orig->m_clones.push_back(s);
        return s;
        }
    case REGEX_SYMBOL:
        {
        Regex *s = m_g->FindOrAddRx(orig->m_Name, ((Regex*)orig)->m_rx, orig->IsCaseInsensitive(), ts, true);
        s->m_cloneof = orig;
        orig->m_clones.push_back(s);
        return s;
        }
    case CHARSET_SYMBOL:
        return orig;
    case NONTERMINAL_SYMBOL:
        {
        Nonterminal *s = m_g->FindOrAddNonterminal(orig->m_Name, ts);
        s->m_cloneof = orig;
        orig->m_clones.push_back(s);
        return s;
        }
    case TERMINAL_SYMBOL:
        return orig;
    default:
        break;
    }
    throw MergeTokenSetStatesException();
    return 0;
}

void TokenSetMergeAlgorithm::Apply()
{
    // Clone each of the symbols that need to be cloned.  When cloning, we
    // reassign the stateless symbol to be the first state carrying symbol,
    // unless it will be used as a nondeterministic token also.  The rest
    // are copies.
    sortedset<Symbol*> reused;
    for( vector<SymbolInState>::iterator cur = m_nondeterministic_tokens.begin(), end = m_nondeterministic_tokens.end(); cur != end; ++cur )
        reused.insert(cur->m_s);

    for( vector<SymbolInState>::iterator cur = m_clones.begin(), end = m_clones.end(); cur != end; ++cur )
    {
        if( reused.find(cur->m_s) == reused.end() )
        {
            // Not reused yet, go ahead and reuse now.
            cur->m_s->m_TokenState = cur->m_state;
            cur->m_clone = cur->m_s;
            reused.insert(cur->m_s);
        }
        else
        {
            // Already reused, clone it.
            cur->m_clone = CloneSymbol(cur->m_s, cur->m_state);
        }
    }
    for( vector<SymbolInState>::iterator cur = m_clones.begin(), end = m_clones.end(); cur != end; ++cur )
    {
        if( cur->m_s->GetSType() != NONTERMINAL_SYMBOL )
            continue;
        Symbol *orignt = cur->m_s;
        Symbol *clonent = cur->m_clone;
        for( vector<Production*>::iterator curp = orignt->m_OwningProductions.begin(), endp = orignt->m_OwningProductions.end(); curp != endp; ++curp )
        {
            Production *origp = *curp;
            ProductionInState pins(origp, cur->m_state);
            vector<ProductionInState>::iterator iter = m_productions.find(pins);
            Production *clonep = CloneProduction(origp, clonent, iter->m_altered_production);
            iter->m_clone = clonep;
        }
    }
    for( sortedset<TokenSetMiss>::iterator cur = m_TokenSetMiss.begin(); cur != m_TokenSetMiss.end(); ++cur )
    {
        vector<ProductionInState>::iterator iter = m_productions.find(cur->m_p);
        cur->m_p.m_p = iter->m_clone;
    }
}

void TokenSetMergeAlgorithm::Report(outbuf *pio)
{
    if( ! pio )
        return;
    // Token state report, but only if there are a variety of token states.
    ob_printf(pio, "Token State Miss Report (%d misses):\n", m_TokenSetMiss.size());
    if( m_TokenSetMiss.size() )
    {
        for( sortedset<TokenSetMiss>::iterator cur = m_TokenSetMiss.begin(); cur != m_TokenSetMiss.end(); ++cur )
            cur->Print(pio);
        ob_putc(pio, '\n');
    }
    else
    {
        ob_puts(pio, "There were no misses.\n\n");
    }
}

TokenSetMiss::TokenSetMiss() : m_pos(-1)
{
}

TokenSetMiss::TokenSetMiss(SymbolInState s, ProductionInState p, int pos)
: m_s(s), m_p(p), m_pos(pos)
{
}

bool TokenSetMiss::operator==(const TokenSetMiss &rhs) const
{
    if( m_s != rhs.m_s || m_p != rhs.m_p || m_pos != rhs.m_pos )
        return false;
    return true;
}

bool TokenSetMiss::operator<(const TokenSetMiss &rhs) const
{
    if( m_s != rhs.m_s )
        return m_s.m_s->GetIndex() < rhs.m_s.m_s->GetIndex();
    if( m_s.m_state != rhs.m_s.m_state )
        return m_s.m_state->GetIndex() < rhs.m_s.m_state->GetIndex();
    if( m_p != rhs.m_p )
        return m_p.m_p->GetIndex() < rhs.m_p.m_p->GetIndex();
    if( m_p.m_state != rhs.m_p.m_state )
        return m_p.m_state->GetIndex() < rhs.m_p.m_state->GetIndex();
    return m_pos < rhs.m_pos;
}

void TokenSetMiss::Print(outbuf *pio)
{
    ob_puts(pio, "The token ");
    m_s.m_s->Print(pio);
    ob_puts(pio, " may need to belong to TokenSet ");
    m_s.m_state->Print(pio);
    ob_puts(pio, ", since it is used at [");
    LRItem pos(0, m_p.m_p, m_pos);
    pos.Print(pio);
    ob_puts(pio, "].\n");
}
